#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.aws.ec2 import subnet as sn
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests import utils

try:
    from neutronclient.common import exceptions as neutron_exc
    from neutronclient.v2_0 import client as neutronclient
except ImportError:
    neutronclient = None


class VPCTestBase(common.HeatTestCase):

    def setUp(self):
        super(VPCTestBase, self).setUp()
        self.mockclient = mock.Mock(spec=neutronclient.Client)
        self.patchobject(neutronclient, 'Client', return_value=self.mockclient)
        self.mockclient.add_interface_router.return_value = None
        self.mockclient.delete_network.return_value = None
        self.mockclient.delete_port.return_value = None
        self.mockclient.delete_router.return_value = None
        self.mockclient.delete_subnet.return_value = None
        self.mockclient.remove_interface_router.return_value = None
        self.mockclient.remove_gateway_router.return_value = None
        self.mockclient.delete_security_group_rule.return_value = None
        self.mockclient.delete_security_group.return_value = None
        self.vpc_name = utils.PhysName('test_stack', 'the_vpc')
        self.mock_router_for_vpc()

        self.subnet_name = utils.PhysName('test_stack', 'the_subnet')
        self.mock_show_subnet()
        self.stub_SubnetConstraint_validate()
        self.mock_create_subnet()

    def create_stack(self, templ):
        t = template_format.parse(templ)
        stack = self.parse_stack(t)
        self.assertIsNone(stack.validate())
        self.assertIsNone(stack.create())
        return stack

    def parse_stack(self, t):
        stack_name = 'test_stack'
        tmpl = template.Template(t)
        stack = parser.Stack(utils.dummy_context(), stack_name, tmpl)
        stack.store()
        return stack

    def validate_mock_create_network(self):
        self.mockclient.show_network.assert_called_with('aaaa')
        self.mockclient.create_network.assert_called_once_with({
            'network': {'name': self.vpc_name}})
        self.mockclient.create_router.assert_called_once()

    def mock_create_network(self):
        self.mockclient.create_network.return_value = {
            'network': {
                'status': 'BUILD',
                'subnets': [],
                'name': 'name',
                'admin_state_up': True,
                'shared': False,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'aaaa'
            }}
        show_network_returns = [
            {"network": {
                "status": "BUILD",
                "subnets": [],
                "name": self.vpc_name,
                "admin_state_up": False,
                "shared": False,
                "tenant_id": "c1210485b2424d48804aad5d39c61b8f",
                "id": "aaaa"
            }}]
        for i in range(7):
            show_network_returns.append(
                {"network": {
                    "status": "ACTIVE",
                    "subnets": [],
                    "name": self.vpc_name,
                    "admin_state_up": False,
                    "shared": False,
                    "tenant_id": "c1210485b2424d48804aad5d39c61b8f",
                    "id": "aaaa"
                }})
        self.mockclient.show_network.side_effect = show_network_returns

        self.mockclient.create_router.return_value = {
            'router': {
                'status': 'BUILD',
                'name': self.vpc_name,
                'admin_state_up': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'bbbb'
            }}

    def mock_create_subnet(self):
        self.subnet_name = utils.PhysName('test_stack', 'the_subnet')
        self.mockclient.create_subnet.return_value = {
            'subnet': {
                'status': 'ACTIVE',
                'name': self.subnet_name,
                'admin_state_up': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'cccc'}}

    def mock_show_subnet(self):
        self.mockclient.show_subnet.return_value = {
            'subnet': {
                'name': self.subnet_name,
                'network_id': 'aaaa',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'allocation_pools': [{'start': '10.0.0.2',
                                      'end': '10.0.0.254'}],
                'gateway_ip': '10.0.0.1',
                'ip_version': 4,
                'cidr': '10.0.0.0/24',
                'id': 'cccc',
                'enable_dhcp': False,
            }}

    def mock_create_security_group(self):
        self.sg_name = utils.PhysName('test_stack', 'the_sg')
        self.mockclient.create_security_group.return_value = {
            'security_group': {
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'name': self.sg_name,
                'description': 'SSH access',
                'security_group_rules': [],
                'id': '0389f747-7785-4757-b7bb-2ab07e4b09c3'
            }
        }
        self.create_security_group_rule_expected = {
            'security_group_rule': {
                'direction': 'ingress',
                'remote_group_id': None,
                'remote_ip_prefix': '0.0.0.0/0',
                'port_range_min': 22,
                'ethertype': 'IPv4',
                'port_range_max': 22,
                'protocol': 'tcp',
                'security_group_id': '0389f747-7785-4757-b7bb-2ab07e4b09c3'
            }
        }

        self.mockclient.create_security_group_rule.return_value = {
            'security_group_rule': {
                'direction': 'ingress',
                'remote_group_id': None,
                'remote_ip_prefix': '0.0.0.0/0',
                'port_range_min': 22,
                'ethertype': 'IPv4',
                'port_range_max': 22,
                'protocol': 'tcp',
                'security_group_id': '0389f747-7785-4757-b7bb-2ab07e4b09c3',
                'id': 'bbbb'
            }
        }

    def mock_show_security_group(self):
        sg_name = utils.PhysName('test_stack', 'the_sg')
        self._group = '0389f747-7785-4757-b7bb-2ab07e4b09c3'
        self.mockclient.show_security_group.return_value = {
            'security_group': {
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'name': sg_name,
                'description': '',
                'security_group_rules': [{
                    'direction': 'ingress',
                    'protocol': 'tcp',
                    'port_range_max': 22,
                    'id': 'bbbb',
                    'ethertype': 'IPv4',
                    'security_group_id': ('0389f747-7785-4757-b7bb-'
                                          '2ab07e4b09c3'),
                    'remote_group_id': None,
                    'remote_ip_prefix': '0.0.0.0/0',
                    'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                    'port_range_min': 22
                }],
                'id': '0389f747-7785-4757-b7bb-2ab07e4b09c3'}}

    def mock_router_for_vpc(self):
        self.mockclient.list_routers.return_value = {
            "routers": [{
                "status": "ACTIVE",
                "external_gateway_info": {
                    "network_id": "zzzz",
                    "enable_snat": True},
                "name": self.vpc_name,
                "admin_state_up": True,
                "tenant_id": "3e21026f2dc94372b105808c0e721661",
                "routes": [],
                "id": "bbbb"
            }]
        }

    def mock_create_route_table(self):
        self.rt_name = utils.PhysName('test_stack', 'the_route_table')
        self.mockclient.create_router.return_value = {
            'router': {
                'status': 'BUILD',
                'name': self.rt_name,
                'admin_state_up': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'ffff'
            }
        }
        show_router_returns = [{
            'router': {
                'status': 'BUILD',
                'name': self.rt_name,
                'admin_state_up': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'ffff'
            }
        }]
        for i in range(3):
            show_router_returns.append({
                'router': {
                    'status': 'ACTIVE',
                    'name': self.rt_name,
                    'admin_state_up': True,
                    'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                    'id': 'ffff'
                }
            })
        self.mockclient.show_router.side_effect = show_router_returns

    def assertResourceState(self, resource, ref_id):
        self.assertIsNone(resource.validate())
        self.assertEqual((resource.CREATE, resource.COMPLETE), resource.state)
        self.assertEqual(ref_id, resource.FnGetRefId())


class VPCTest(VPCTestBase):

    test_template = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
'''

    def mock_create_network_failed(self):
        self.vpc_name = utils.PhysName('test_stack', 'the_vpc')
        exc = neutron_exc.NeutronClientException
        self.mockclient.create_network.side_effect = exc

    def test_vpc(self):
        self.mock_create_network()

        stack = self.create_stack(self.test_template)
        vpc = stack['the_vpc']
        self.assertResourceState(vpc, 'aaaa')
        self.validate_mock_create_network()
        self.assertEqual(3, self.mockclient.show_network.call_count)

        scheduler.TaskRunner(vpc.delete)()

        self.mockclient.show_network.assert_called_with('aaaa')
        self.assertEqual(4, self.mockclient.show_network.call_count)
        self.assertEqual(2, self.mockclient.list_routers.call_count)
        self.mockclient.list_routers.assert_called_with(name=self.vpc_name)
        self.mockclient.delete_router.assert_called_once_with('bbbb')
        self.mockclient.delete_network.assert_called_once_with('aaaa')

    def test_vpc_delete_successful_if_created_failed(self):
        self.mock_create_network_failed()

        t = template_format.parse(self.test_template)
        stack = self.parse_stack(t)
        scheduler.TaskRunner(stack.create)()
        self.assertEqual((stack.CREATE, stack.FAILED), stack.state)
        self.mockclient.create_network.assert_called_once_with(
            {'network': {'name': self.vpc_name}})
        scheduler.TaskRunner(stack.delete)()
        self.mockclient.delete_network.assert_not_called()


class SubnetTest(VPCTestBase):

    test_template = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
'''

    def test_subnet(self):
        self.mock_create_network()
        exc = neutron_exc.NeutronClientException(status_code=404)
        self.mockclient.remove_interface_router.side_effect = exc
        self.mockclient.delete_subnet.side_effect = exc

        stack = self.create_stack(self.test_template)

        subnet = stack['the_subnet']

        self.assertResourceState(subnet, 'cccc')
        self.mockclient.list_routers.assert_called_with(name=self.vpc_name)

        self.validate_mock_create_network()

        self.mockclient.add_interface_router.assert_called_once_with(
            'bbbb', {'subnet_id': 'cccc'})
        self.mockclient.create_subnet.assert_called_once_with(
            {'subnet': {
                'network_id': 'aaaa',
                'cidr': '10.0.0.0/24',
                'ip_version': 4,
                'name': self.subnet_name}})
        self.assertEqual(4, self.mockclient.show_network.call_count)
        self.assertRaises(
            exception.InvalidTemplateAttribute,
            subnet.FnGetAtt,
            'Foo')
        self.assertEqual('moon', subnet.FnGetAtt('AvailabilityZone'))

        scheduler.TaskRunner(subnet.delete)()
        subnet.state_set(subnet.CREATE, subnet.COMPLETE, 'to delete again')
        scheduler.TaskRunner(subnet.delete)()
        scheduler.TaskRunner(stack['the_vpc'].delete)()

        self.mockclient.show_network.assert_called_with('aaaa')
        self.mockclient.list_routers.assert_called_with(name=self.vpc_name)
        self.assertEqual(5, self.mockclient.list_routers.call_count)

        self.assertEqual(7, self.mockclient.show_network.call_count)

    def _mock_create_subnet_failed(self, stack_name):
        self.subnet_name = utils.PhysName(stack_name, 'the_subnet')
        self.mockclient.create_subnet.return_value = {
            'subnet': {
                'status': 'ACTIVE',
                'name': self.subnet_name,
                'admin_state_up': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'id': 'cccc'}}
        exc = neutron_exc.NeutronClientException(
            status_code=404)
        self.mockclient.show_network.side_effect = exc

    def test_create_failed_delete_success(self):
        stack_name = 'test_subnet_'
        self._mock_create_subnet_failed(stack_name)

        t = template_format.parse(self.test_template)
        tmpl = template.Template(t)
        stack = parser.Stack(utils.dummy_context(), stack_name, tmpl,
                             stack_id=str(uuid.uuid4()))
        tmpl.t['Resources']['the_subnet']['Properties']['VpcId'] = 'aaaa'
        resource_defns = tmpl.resource_definitions(stack)
        rsrc = sn.Subnet('the_subnet',
                         resource_defns['the_subnet'],
                         stack)
        rsrc.validate()
        self.assertRaises(exception.ResourceFailure,
                          scheduler.TaskRunner(rsrc.create))
        self.assertEqual((rsrc.CREATE, rsrc.FAILED), rsrc.state)
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('cccc', ref_id)

        self.mockclient.create_subnet.assert_called_once_with(
            {'subnet': {
                'network_id': 'aaaa',
                'cidr': '10.0.0.0/24',
                'ip_version': 4,
                'name': self.subnet_name}})
        self.assertEqual(1, self.mockclient.show_network.call_count)

        self.assertIsNone(scheduler.TaskRunner(rsrc.delete)())
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.assertEqual(2, self.mockclient.show_network.call_count)
        self.mockclient.delete_subnet.assert_called_once_with('cccc')


class NetworkInterfaceTest(VPCTestBase):

    test_template = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_sg:
    Type: AWS::EC2::SecurityGroup
    Properties:
      VpcId: {Ref: the_vpc}
      GroupDescription: SSH access
      SecurityGroupIngress:
        - IpProtocol: tcp
          FromPort: "22"
          ToPort: "22"
          CidrIp: 0.0.0.0/0
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_nic:
    Type: AWS::EC2::NetworkInterface
    Properties:
      PrivateIpAddress: 10.0.0.100
      SubnetId: {Ref: the_subnet}
      GroupSet:
      - Ref: the_sg
'''

    test_template_no_groupset = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_nic:
    Type: AWS::EC2::NetworkInterface
    Properties:
      PrivateIpAddress: 10.0.0.100
      SubnetId: {Ref: the_subnet}
'''

    test_template_error = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_sg:
    Type: AWS::EC2::SecurityGroup
    Properties:
      VpcId: {Ref: the_vpc}
      GroupDescription: SSH access
      SecurityGroupIngress:
        - IpProtocol: tcp
          FromPort: "22"
          ToPort: "22"
          CidrIp: 0.0.0.0/0
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_nic:
    Type: AWS::EC2::NetworkInterface
    Properties:
      PrivateIpAddress: 10.0.0.100
      SubnetId: {Ref: the_subnet}
      GroupSet:
      - Ref: INVALID-REF-IN-TEMPLATE
'''

    test_template_error_no_ref = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_vpc:
    Type: AWS::EC2::VPC
    Properties: {CidrBlock: '10.0.0.0/16'}
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_nic:
    Type: AWS::EC2::NetworkInterface
    Properties:
      PrivateIpAddress: 10.0.0.100
      SubnetId: {Ref: the_subnet}
      GroupSet:
      - INVALID-NO-REF
'''

    def setUp(self):
        super(NetworkInterfaceTest, self).setUp()
        self.mock_create_network()
        self.mock_create_security_group()
        self.mock_show_security_group()
        self.mock_show_network_interface()

    def mock_create_network_interface(
            self, security_groups=['0389f747-7785-4757-b7bb-2ab07e4b09c3']):

        self.nic_name = utils.PhysName('test_stack', 'the_nic')
        self._port = {'network_id': 'aaaa',
                      'fixed_ips': [{
                          'subnet_id': 'cccc',
                          'ip_address': '10.0.0.100'
                      }],
                      'name': self.nic_name,
                      'admin_state_up': True}
        if security_groups:
            self._port['security_groups'] = security_groups

        self.mockclient.create_port.return_value = {
            'port': {
                'admin_state_up': True,
                'device_id': '',
                'device_owner': '',
                'fixed_ips': [
                    {
                        'ip_address': '10.0.0.100',
                        'subnet_id': 'cccc'
                    }
                ],
                'id': 'dddd',
                'mac_address': 'fa:16:3e:25:32:5d',
                'name': self.nic_name,
                'network_id': 'aaaa',
                'status': 'ACTIVE',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f'
            }
        }

    def mock_show_network_interface(self):
        self.nic_name = utils.PhysName('test_stack', 'the_nic')
        self.mockclient.show_port.return_value = {
            'port': {
                'admin_state_up': True,
                'device_id': '',
                'device_owner': '',
                'fixed_ips': [
                    {
                        'ip_address': '10.0.0.100',
                        'subnet_id': 'cccc'
                    }
                ],
                'id': 'dddd',
                'mac_address': 'fa:16:3e:25:32:5d',
                'name': self.nic_name,
                'network_id': 'aaaa',
                'security_groups': ['0389f747-7785-4757-b7bb-2ab07e4b09c3'],
                'status': 'ACTIVE',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f'
            }
        }

    def test_network_interface(self):
        self.mock_create_network_interface()

        stack = self.create_stack(self.test_template)
        self.validate_mock_create_network()
        self.mockclient.create_port.assert_called_once_with(
            {'port': self._port})
        self.mockclient.create_security_group.assert_called_once_with({
            'security_group': {
                'name': self.sg_name,
                'description': 'SSH access'
            }
        })
        self.mockclient.create_security_group_rule.assert_called_once_with(
            self.create_security_group_rule_expected)
        try:
            self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)
            rsrc = stack['the_nic']
            self.assertResourceState(rsrc, 'dddd')
            self.assertEqual('10.0.0.100', rsrc.FnGetAtt('PrivateIpAddress'))
        finally:
            scheduler.TaskRunner(stack.delete)()
        self.mockclient.delete_port.assert_called_once_with('dddd')
        self.mockclient.show_subnet.assert_called_once_with('cccc')
        self.mockclient.show_port.assert_called_once_with('dddd')
        self.mockclient.show_security_group.assert_called_once_with(
            self._group)
        self.mockclient.delete_security_group.assert_called_once_with(
            '0389f747-7785-4757-b7bb-2ab07e4b09c3')
        self.mockclient.delete_security_group_rule.assert_called_once_with(
            'bbbb')

    def test_network_interface_existing_groupset(self):
        self.patchobject(parser.Stack, 'resource_by_refid')

        self.mock_create_network_interface()

        stack = self.create_stack(self.test_template)
        self.validate_mock_create_network()
        self.mockclient.create_port.assert_called_once_with(
            {'port': self._port})
        self.mockclient.create_security_group.assert_called_once_with({
            'security_group': {
                'name': self.sg_name,
                'description': 'SSH access'
            }
        })
        self.mockclient.create_security_group_rule.assert_called_once_with(
            self.create_security_group_rule_expected)
        try:
            self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)
            rsrc = stack['the_nic']
            self.assertResourceState(rsrc, 'dddd')
        finally:
            stack.delete()
        self.mockclient.delete_port.assert_called_once_with('dddd')
        self.mockclient.show_subnet.assert_called_once_with('cccc')
        self.mockclient.show_security_group.assert_called_once_with(
            self._group)
        self.mockclient.delete_security_group.assert_called_once_with(
            '0389f747-7785-4757-b7bb-2ab07e4b09c3')
        self.mockclient.delete_security_group_rule.assert_called_once_with(
            'bbbb')

    def test_network_interface_no_groupset(self):
        self.mock_create_network_interface(security_groups=None)

        stack = self.create_stack(self.test_template_no_groupset)
        self.mockclient.create_port.assert_called_once_with(
            {'port': self._port})
        stack.delete()
        self.mockclient.delete_port.assert_called_once_with('dddd')
        self.mockclient.show_subnet.assert_called_once_with('cccc')

    def test_network_interface_error(self):
        self.assertRaises(
            exception.StackValidationFailed,
            self.create_stack,
            self.test_template_error)


class InternetGatewayTest(VPCTestBase):

    test_template = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_gateway:
    Type: AWS::EC2::InternetGateway
  the_vpc:
    Type: AWS::EC2::VPC
    Properties:
      CidrBlock: '10.0.0.0/16'
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_attachment:
    Type: AWS::EC2::VPCGatewayAttachment
    Properties:
      VpcId: {Ref: the_vpc}
      InternetGatewayId: {Ref: the_gateway}
  the_route_table:
    Type: AWS::EC2::RouteTable
    Properties:
      VpcId: {Ref: the_vpc}
  the_association:
    Type: AWS::EC2::SubnetRouteTableAssociation
    Properties:
      RouteTableId: {Ref: the_route_table}
      SubnetId: {Ref: the_subnet}
'''

    def setUp(self):
        super(InternetGatewayTest, self).setUp()
        self.mock_create_internet_gateway()
        self.mock_create_network()
        self.mock_create_route_table()

    def mock_create_internet_gateway(self):
        self.mockclient.list_networks.return_value = {
            'networks': [{
                'status': 'ACTIVE',
                'subnets': [],
                'name': 'nova',
                'router:external': True,
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'admin_state_up': True,
                'shared': True,
                'id': '0389f747-7785-4757-b7bb-2ab07e4b09c3'
            }]}

    def test_internet_gateway(self):

        stack = self.create_stack(self.test_template)
        self.mockclient.show_network.assert_called_with('aaaa')
        self.mockclient.create_network.assert_called_with({
            'network': {'name': self.vpc_name}})
        self.assertEqual(2, self.mockclient.create_router.call_count)
        self.mockclient.create_router.assert_called_with(
            {'router': {'name': self.rt_name}})
        self.mockclient.add_interface_router.assert_has_calls([
            mock.call('bbbb', {'subnet_id': 'cccc'}),
            mock.call('ffff', {'subnet_id': 'cccc'}),
        ])
        self.mockclient.list_networks.assert_called_once_with(
            **{'router:external': True})

        gateway = stack['the_gateway']
        self.assertResourceState(gateway, gateway.physical_resource_name())
        self.mockclient.add_gateway_router.assert_called_with(
            'ffff', {'network_id': '0389f747-7785-4757-b7bb-2ab07e4b09c3'})

        attachment = stack['the_attachment']
        self.assertResourceState(attachment, 'the_attachment')

        route_table = stack['the_route_table']
        self.assertEqual([route_table], list(attachment._vpc_route_tables()))

        stack.delete()
        self.mockclient.remove_interface_router.assert_has_calls([
            mock.call('ffff', {'subnet_id': 'cccc'}),
            mock.call('bbbb', {'subnet_id': 'cccc'}),
        ])
        self.mockclient.remove_gateway_router.assert_called_with('ffff')
        self.assertEqual(2, self.mockclient.remove_gateway_router.call_count)
        self.assertEqual(2, self.mockclient.show_subnet.call_count)
        self.mockclient.show_subnet.assert_called_with('cccc')
        self.mockclient.show_router.assert_called_with('ffff')
        self.assertEqual(2, self.mockclient.show_router.call_count)
        self.mockclient.delete_router.assert_called_once_with('ffff')


class RouteTableTest(VPCTestBase):

    test_template = '''
HeatTemplateFormatVersion: '2012-12-12'
Resources:
  the_vpc:
    Type: AWS::EC2::VPC
    Properties:
      CidrBlock: '10.0.0.0/16'
  the_subnet:
    Type: AWS::EC2::Subnet
    Properties:
      CidrBlock: 10.0.0.0/24
      VpcId: {Ref: the_vpc}
      AvailabilityZone: moon
  the_route_table:
    Type: AWS::EC2::RouteTable
    Properties:
      VpcId: {Ref: the_vpc}
  the_association:
    Type: AWS::EC2::SubnetRouteTableAssociation
    Properties:
      RouteTableId: {Ref: the_route_table}
      SubnetId: {Ref: the_subnet}
'''

    def setUp(self):
        super(RouteTableTest, self).setUp()
        self.mock_create_network()
        self.mock_create_route_table()

    def test_route_table(self):
        stack = self.create_stack(self.test_template)
        self.mockclient.create_router.assert_called_with(
            {'router': {'name': self.rt_name}})
        self.mockclient.add_interface_router.assert_has_calls([
            mock.call('bbbb', {'subnet_id': 'cccc'}),
            mock.call('ffff', {'subnet_id': 'cccc'}),
        ])

        route_table = stack['the_route_table']
        self.assertResourceState(route_table, 'ffff')
        self.mockclient.show_network.assert_called_with('aaaa')
        self.mockclient.create_network.assert_called_with({
            'network': {'name': self.vpc_name}})
        self.assertEqual(2, self.mockclient.create_router.call_count)

        association = stack['the_association']
        self.assertResourceState(association, 'the_association')

        scheduler.TaskRunner(association.delete)()
        scheduler.TaskRunner(route_table.delete)()

        stack.delete()
        self.mockclient.remove_interface_router.assert_has_calls([
            mock.call('ffff', {'subnet_id': 'cccc'}),
            mock.call('bbbb', {'subnet_id': 'cccc'}),
        ])
        self.mockclient.remove_gateway_router.assert_called_once_with('ffff')
        self.assertEqual(2, self.mockclient.show_subnet.call_count)
        self.mockclient.show_subnet.assert_called_with('cccc')
        self.assertEqual(2, self.mockclient.show_router.call_count)
        self.mockclient.show_router.assert_called_with('ffff')
        self.mockclient.delete_router.assert_called_once_with('ffff')



#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine.clients.os import neutron
from heat.engine.resources.openstack.neutron.sfc import port_chain
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

port_chain_template = {
    'heat_template_version': '2015-04-30',
    'resources': {
        'test_resource': {
            'type': 'OS::Neutron::PortChain',
            'properties': {
                'name': 'test_port_chain',
                'description': 'port_chain_desc',
                'port_pair_groups': ['port_pair_group_1'],
                'flow_classifiers': ['flow_classifier1'],
                'chain_parameters': {"correlation": 'mpls'}
                }
            }
        }
    }


class PortChainTest(common.HeatTestCase):

    def setUp(self):
        super(PortChainTest, self).setUp()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(port_chain_template)
        )
        self.test_resource = self.stack['test_resource']

        self.test_client_plugin = mock.MagicMock()

        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

        self.test_client_plugin.get_notification = mock.MagicMock(
            return_value='sample_notification')

        self.patchobject(self.test_client_plugin, 'resolve_ext_resource'
                         ).return_value = ('port_pair_group_1')

        self.patchobject(self.test_client_plugin, 'resolve_ext_resource'
                         ).return_value = ('flow_classifier1')

    def test_resource_mapping(self):
        mapping = port_chain.resource_mapping()
        self.assertEqual(port_chain.PortChain,
                         mapping['OS::Neutron::PortChain'])

    def _get_mock_resource(self):
        value = mock.MagicMock()
        value.id = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        return value

    def test_resource_handle_create(self):
        mock_pc_create = self.test_client_plugin.create_ext_resource
        mock_resource = self._get_mock_resource()
        mock_pc_create.return_value = mock_resource

    # validate the properties
        self.assertEqual(
            'test_port_chain',
            self.test_resource.properties.get(
                port_chain.PortChain.NAME))
        self.assertEqual(
            'port_chain_desc',
            self.test_resource.properties.get(
                port_chain.PortChain.DESCRIPTION))
        self.assertEqual(
            ['port_pair_group_1'],
            self.test_resource.properties.get(
                port_chain.PortChain.PORT_PAIR_GROUPS))
        self.assertEqual(
            ['flow_classifier1'],
            self.test_resource.properties.get(
                port_chain.PortChain.FLOW_CLASSIFIERS))
        self.assertEqual(
            {"correlation": 'mpls'},
            self.test_resource.properties.get(
                port_chain.PortChain.CHAIN_PARAMETERS))

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        mock_pc_create.assert_called_once_with(
            'port_chain',
            {
                'name': 'test_port_chain',
                'description': 'port_chain_desc',
                'port_pair_groups': ['port_pair_group_1'],
                'flow_classifiers': ['flow_classifier1'],
                'chain_parameters': {"correlation": 'mpls'}}
        )

    def delete_portchain(self):
        mock_pc_delete = self.test_client_plugin.delete_ext_resource
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_pc_delete.return_value = None
        self.assertIsNone(self.test_resource.handle_delete())
        mock_pc_delete.assert_called_once_with(
            'port_chain', self.test_resource.resource_id)

    def delete_portchain_resource_id_is_none(self):
        self.test_resource.resource_id = None
        self.assertIsNone(self.test_resource.handle_delete())
        self.assertEqual(0, self.test_client_plugin.
                         delete_ext_resource.call_count)

    def test_resource_handle_delete_not_found(self):
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_pc_delete = self.test_client_plugin.delete_ext_resource
        mock_pc_delete.side_effect = self.test_client_plugin.NotFound
        self.assertIsNone(self.test_resource.handle_delete())

    def test_resource_show_resource(self):
        mock_pc_get = self.test_client_plugin.show_ext_resource
        mock_pc_get.return_value = None
        self.assertIsNone(self.test_resource._show_resource(),
                          'Failed to show resource')

    def test_resource_handle_update(self):
        mock_ppg_patch = self.test_client_plugin.update_ext_resource
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        prop_diff = {
            'name': 'name-updated',
            'description': 'description-updated',
            'port_pair_groups': ['port_pair_group_2'],
            'flow_classifiers': ['flow_classifier2'],
        }
        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        mock_ppg_patch.assert_called_once_with(
            'port_chain',
            {
                'name': 'name-updated',
                'description': 'description-updated',
                'port_pair_groups': ['port_pair_group_2'],
                'flow_classifiers': ['flow_classifier2'],
            }, self.test_resource.resource_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support

DEPR_MSG = _('Neutron LBaaS v1 is deprecated in the Liberty release '
             'and is planned to be removed in a future release. '
             'Going forward, the LBaaS V2 should be used.')


class HealthMonitor(none_resource.NoneResource):
    """A resource for managing health monitors for loadbalancers in Neutron.

    A health monitor is used to determine whether or not back-end members of
    the VIP's pool are usable for processing a request. A pool can have several
    health monitors associated with it. There are different types of health
    monitors supported by the OpenStack LBaaS service:

      - PING: used to ping the members using ICMP.
      - TCP: used to connect to the members using TCP.
      - HTTP: used to send an HTTP request to the member.
      - HTTPS: used to send a secure HTTP request to the member.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='9.0.0',
        message=_('Use LBaaS V2 instead.'),
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=DEPR_MSG,
            version='7.0.0'
        )
    )


class Pool(none_resource.NoneResource):
    """A resource for managing load balancer pools in Neutron.

    A load balancing pool is a logical set of devices, such as web servers,
    that you group together to receive and process traffic. The loadbalancing
    function chooses a member of the pool according to the configured load
    balancing method to handle the new requests or connections received on the
    VIP address. There is only one pool for a VIP.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='9.0.0',
        message=_('Use LBaaS V2 instead.'),
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=DEPR_MSG,
            version='7.0.0'
        )
    )


class PoolMember(none_resource.NoneResource):
    """A resource to handle loadbalancer members.

    A pool member represents the application running on backend server.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='9.0.0',
        message=_('Use LBaaS V2 instead.'),
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=DEPR_MSG,
            version='7.0.0',
            previous_status=support.SupportStatus(version='2014.1')
        )
    )


class LoadBalancer(none_resource.NoneResource):
    """A resource to link a neutron pool with servers.

    A loadbalancer allows linking a neutron pool with specified servers to some
    port.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='9.0.0',
        message=_('Use LBaaS V2 instead.'),
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=DEPR_MSG,
            version='7.0.0',
            previous_status=support.SupportStatus(version='2014.1')
        )
    )


def resource_mapping():
    return {
        'OS::Neutron::HealthMonitor': HealthMonitor,
        'OS::Neutron::Pool': Pool,
        'OS::Neutron::PoolMember': PoolMember,
        'OS::Neutron::LoadBalancer': LoadBalancer,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

import fixtures
from oslo_config import cfg
from oslo_log import log as logging
import testscenarios
import testtools

from heat.common import context
from heat.common import messaging
from heat.common import policy
from heat.engine.clients.os import barbican
from heat.engine.clients.os import cinder
from heat.engine.clients.os import glance
from heat.engine.clients.os import keystone
from heat.engine.clients.os.keystone import fake_keystoneclient as fake_ks
from heat.engine.clients.os.keystone import keystone_constraints as ks_constr
from heat.engine.clients.os.neutron import neutron_constraints as neutron
from heat.engine.clients.os import nova
from heat.engine.clients.os import trove
from heat.engine import environment
from heat.engine import resource
from heat.engine import resources
from heat.engine import scheduler
from heat.tests import fakes
from heat.tests import generic_resource as generic_rsrc
from heat.tests import utils

TEST_DEFAULT_LOGLEVELS = {'migrate': logging.WARN,
                          'sqlalchemy': logging.WARN,
                          'heat.engine.environment': logging.ERROR}
_LOG_FORMAT = "%(levelname)8s [%(name)s] %(message)s"
_TRUE_VALUES = ('True', 'true', '1', 'yes')


class FakeLogMixin(object):
    def setup_logging(self, quieten=True):
        # Assign default logs to self.LOG so we can still
        # assert on heat logs.
        default_level = logging.INFO
        if os.environ.get('OS_DEBUG') in _TRUE_VALUES:
            default_level = logging.DEBUG

        self.LOG = self.useFixture(
            fixtures.FakeLogger(level=default_level, format=_LOG_FORMAT))
        base_list = set([nlog.split('.')[0]
                         for nlog in logging.logging.Logger.manager.loggerDict]
                        )
        for base in base_list:
            if base in TEST_DEFAULT_LOGLEVELS:
                self.useFixture(fixtures.FakeLogger(
                    level=TEST_DEFAULT_LOGLEVELS[base],
                    name=base, format=_LOG_FORMAT))
            elif base != 'heat':
                self.useFixture(fixtures.FakeLogger(
                    name=base, format=_LOG_FORMAT))
        if quieten:
            for ll in TEST_DEFAULT_LOGLEVELS:
                if ll.startswith('heat.'):
                    self.useFixture(fixtures.FakeLogger(
                        level=TEST_DEFAULT_LOGLEVELS[ll],
                        name=ll, format=_LOG_FORMAT))


class HeatTestCase(testscenarios.WithScenarios,
                   testtools.TestCase, FakeLogMixin):

    def setUp(self, mock_keystone=True, mock_resource_policy=True,
              quieten_logging=True, mock_find_file=True):
        super().setUp()
        self.setup_logging(quieten=quieten_logging)

        self.useFixture(utils.WarningsFixture())

        scheduler.ENABLE_SLEEP = False
        self.useFixture(fixtures.MonkeyPatch(
            'heat.common.exception._FATAL_EXCEPTION_FORMAT_ERRORS',
            True))

        def enable_sleep():
            scheduler.ENABLE_SLEEP = True

        self.addCleanup(enable_sleep)

        mod_dir = os.path.dirname(sys.modules[__name__].__file__)
        project_dir = os.path.abspath(os.path.join(mod_dir, '../../'))
        env_dir = os.path.join(project_dir, 'etc', 'heat',
                               'environment.d')
        template_dir = os.path.join(project_dir, 'etc', 'heat',
                                    'templates')

        cfg.CONF.set_default('environment_dir', env_dir)
        cfg.CONF.set_override('error_wait_time', None)
        cfg.CONF.set_default('template_dir', template_dir)
        self.addCleanup(cfg.CONF.reset)

        messaging.setup("fake://", optional=True)
        self.addCleanup(messaging.cleanup)

        tri_names = ['AWS::RDS::DBInstance', 'AWS::CloudWatch::Alarm']
        tris = []
        for name in tri_names:
            tris.append(resources.global_env().get_resource_info(
                name,
                registry_type=environment.TemplateResourceInfo))
        for tri in tris:
            if tri is not None:
                cur_path = tri.template_name
                templ_path = os.path.join(project_dir, 'etc',
                                          'heat', 'templates')
                if templ_path not in cur_path:
                    tri.template_name = cur_path.replace(
                        '/etc/heat/templates',
                        templ_path)

        if mock_find_file:
            self.mock_find_file = self.patchobject(
                cfg.ConfigOpts, 'find_file')
        if mock_keystone:
            self.stub_keystoneclient()
        if mock_resource_policy:
            self.mock_resource_policy = self.patchobject(
                policy.ResourceEnforcer, 'enforce')
        utils.setup_dummy_db()
        self.register_test_resources()
        self.addCleanup(utils.reset_dummy_db)

    def register_test_resources(self):
        resource._register_class('GenericResourceType',
                                 generic_rsrc.GenericResource)
        resource._register_class('MultiStepResourceType',
                                 generic_rsrc.MultiStepResource)
        resource._register_class('ResWithShowAttrType',
                                 generic_rsrc.ResWithShowAttr)
        resource._register_class('SignalResourceType',
                                 generic_rsrc.SignalResource)
        resource._register_class('ResourceWithPropsType',
                                 generic_rsrc.ResourceWithProps)
        resource._register_class('ResourceWithPropsRefPropOnDelete',
                                 generic_rsrc.ResourceWithPropsRefPropOnDelete)
        resource._register_class(
            'ResourceWithPropsRefPropOnValidate',
            generic_rsrc.ResourceWithPropsRefPropOnValidate)
        resource._register_class('StackUserResourceType',
                                 generic_rsrc.StackUserResource)
        resource._register_class('ResourceWithResourceIDType',
                                 generic_rsrc.ResourceWithResourceID)
        resource._register_class('ResourceWithAttributeType',
                                 generic_rsrc.ResourceWithAttributeType)
        resource._register_class('ResourceWithRequiredProps',
                                 generic_rsrc.ResourceWithRequiredProps)
        resource._register_class(
            'ResourceWithMultipleRequiredProps',
            generic_rsrc.ResourceWithMultipleRequiredProps)
        resource._register_class(
            'ResourceWithRequiredPropsAndEmptyAttrs',
            generic_rsrc.ResourceWithRequiredPropsAndEmptyAttrs)
        resource._register_class('ResourceWithPropsAndAttrs',
                                 generic_rsrc.ResourceWithPropsAndAttrs)
        resource._register_class('ResWithStringPropAndAttr',
                                 generic_rsrc.ResWithStringPropAndAttr),
        resource._register_class('ResWithComplexPropsAndAttrs',
                                 generic_rsrc.ResWithComplexPropsAndAttrs)
        resource._register_class('ResourceWithCustomConstraint',
                                 generic_rsrc.ResourceWithCustomConstraint)
        resource._register_class('ResourceWithComplexAttributesType',
                                 generic_rsrc.ResourceWithComplexAttributes)
        resource._register_class('ResourceWithDefaultClientName',
                                 generic_rsrc.ResourceWithDefaultClientName)
        resource._register_class('OverwrittenFnGetAttType',
                                 generic_rsrc.ResourceWithFnGetAttType)
        resource._register_class('OverwrittenFnGetRefIdType',
                                 generic_rsrc.ResourceWithFnGetRefIdType)
        resource._register_class('ResourceWithListProp',
                                 generic_rsrc.ResourceWithListProp)
        resource._register_class('StackResourceType',
                                 generic_rsrc.StackResourceType)
        resource._register_class('ResourceWithRestoreType',
                                 generic_rsrc.ResourceWithRestoreType)
        resource._register_class('ResourceTypeUnSupportedLiberty',
                                 generic_rsrc.ResourceTypeUnSupportedLiberty)
        resource._register_class('ResourceTypeSupportedKilo',
                                 generic_rsrc.ResourceTypeSupportedKilo)
        resource._register_class('ResourceTypeHidden',
                                 generic_rsrc.ResourceTypeHidden)
        resource._register_class(
            'ResourceWithHiddenPropertyAndAttribute',
            generic_rsrc.ResourceWithHiddenPropertyAndAttribute)

    def patchobject(self, obj, attr, **kwargs):
        mockfixture = self.useFixture(fixtures.MockPatchObject(obj, attr,
                                                               **kwargs))
        return mockfixture.mock

    # NOTE(pshchelo): this overrides the testtools.TestCase.patch method
    # that does simple monkey-patching in favor of mock's patching
    def patch(self, target, **kwargs):
        mockfixture = self.useFixture(fixtures.MockPatch(target,
                                                         **kwargs))
        return mockfixture.mock

    def stub_auth(self, ctx=None, **kwargs):
        auth = self.patchobject(ctx or context.RequestContext,
                                "_create_auth_plugin")
        fake_auth = fakes.FakeAuth(**kwargs)
        auth.return_value = fake_auth
        return auth

    def stub_keystoneclient(self, fake_client=None, **kwargs):
        client = self.patchobject(keystone.KeystoneClientPlugin, "_create")
        fkc = fake_client or fake_ks.FakeKeystoneClient(**kwargs)
        client.return_value = fkc
        return fkc

    def stub_KeypairConstraint_validate(self):
        validate = self.patchobject(nova.KeypairConstraint, 'validate')
        validate.return_value = True

    def stub_ImageConstraint_validate(self, num=None):
        validate = self.patchobject(glance.ImageConstraint, 'validate')
        if num is None:
            validate.return_value = True
        else:
            validate.side_effect = [True for x in range(num)]

    def stub_FlavorConstraint_validate(self):
        validate = self.patchobject(nova.FlavorConstraint, 'validate')
        validate.return_value = True

    def stub_VolumeConstraint_validate(self):
        validate = self.patchobject(cinder.VolumeConstraint, 'validate')
        validate.return_value = True

    def stub_QoSSpecsConstraint_validate(self):
        validate = self.patchobject(cinder.QoSSpecsConstraint, 'validate')
        validate.return_value = True

    def stub_SnapshotConstraint_validate(self):
        validate = self.patchobject(
            cinder.VolumeSnapshotConstraint, 'validate')
        validate.return_value = True

    def stub_VolumeTypeConstraint_validate(self):
        validate = self.patchobject(cinder.VolumeTypeConstraint, 'validate')
        validate.return_value = True

    def stub_VolumeBackupConstraint_validate(self):
        validate = self.patchobject(cinder.VolumeBackupConstraint, 'validate')
        validate.return_value = True

    def stub_ServerConstraint_validate(self):
        validate = self.patchobject(nova.ServerConstraint, 'validate')
        validate.return_value = True

    def stub_NetworkConstraint_validate(self):
        validate = self.patchobject(neutron.NetworkConstraint, 'validate')
        validate.return_value = True

    def stub_PortConstraint_validate(self):
        validate = self.patchobject(neutron.PortConstraint, 'validate')
        validate.return_value = True

    def stub_TroveFlavorConstraint_validate(self):
        validate = self.patchobject(trove.FlavorConstraint, 'validate')
        validate.return_value = True

    def stub_SubnetConstraint_validate(self):
        validate = self.patchobject(neutron.SubnetConstraint, 'validate')
        validate.return_value = True

    def stub_AddressScopeConstraint_validate(self):
        validate = self.patchobject(neutron.AddressScopeConstraint, 'validate')
        validate.return_value = True

    def stub_SubnetPoolConstraint_validate(self):
        validate = self.patchobject(neutron.SubnetPoolConstraint, 'validate')
        validate.return_value = True

    def stub_RouterConstraint_validate(self):
        validate = self.patchobject(neutron.RouterConstraint, 'validate')
        validate.return_value = True

    def stub_QoSPolicyConstraint_validate(self):
        validate = self.patchobject(neutron.QoSPolicyConstraint, 'validate')
        validate.return_value = True

    def stub_KeystoneProjectConstraint(self):
        validate = self.patchobject(ks_constr.KeystoneProjectConstraint,
                                    'validate')
        validate.return_value = True

    def stub_ProviderConstraint_validate(self):
        validate = self.patchobject(neutron.ProviderConstraint, 'validate')
        validate.return_value = True

    def stub_SecretConstraint_validate(self):
        validate = self.patchobject(barbican.SecretConstraint, 'validate')
        validate.return_value = True



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for :module:'heat.engine.clients.os.cinder'."""

from unittest import mock
import uuid

from cinderclient import api_versions as cinder_api_versions
from cinderclient import exceptions as cinder_exc
from keystoneauth1 import exceptions as ks_exceptions

from heat.common import exception
from heat.engine.clients.os import cinder
from heat.tests import common
from heat.tests import utils


class CinderClientPluginTest(common.HeatTestCase):
    """Basic tests for :module:'heat.engine.clients.os.cinder'."""

    def setUp(self):
        super(CinderClientPluginTest, self).setUp()
        self.cinder_client = mock.MagicMock()
        con = utils.dummy_context()
        c = con.clients
        self.cinder_plugin = c.client_plugin('cinder')
        self.cinder_plugin.client = lambda: self.cinder_client

    def test_get_volume(self):
        """Tests the get_volume function."""
        volume_id = str(uuid.uuid4())
        my_volume = mock.MagicMock()
        self.cinder_client.volumes.get.return_value = my_volume

        self.assertEqual(my_volume, self.cinder_plugin.get_volume(volume_id))
        self.cinder_client.volumes.get.assert_called_once_with(volume_id)

    def test_get_snapshot(self):
        """Tests the get_volume_snapshot function."""
        snapshot_id = str(uuid.uuid4())
        my_snapshot = mock.MagicMock()
        self.cinder_client.volume_snapshots.get.return_value = my_snapshot

        self.assertEqual(my_snapshot,
                         self.cinder_plugin.get_volume_snapshot(snapshot_id))
        self.cinder_client.volume_snapshots.get.assert_called_once_with(
            snapshot_id)


class VolumeConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(VolumeConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_volume = mock.Mock()
        self.ctx.clients.client_plugin(
            'cinder').get_volume = self.mock_get_volume
        self.constraint = cinder.VolumeConstraint()

    def test_validation(self):
        self.mock_get_volume.return_value = None
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_volume.side_effect = exception.EntityNotFound(
            entity='Volume', name='bar')
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class VolumeSnapshotConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(VolumeSnapshotConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_snapshot = mock.Mock()
        self.ctx.clients.client_plugin(
            'cinder').get_volume_snapshot = self.mock_get_snapshot
        self.constraint = cinder.VolumeSnapshotConstraint()

    def test_validation(self):
        self.mock_get_snapshot.return_value = 'snapshot'
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_snapshot.side_effect = exception.EntityNotFound(
            entity='VolumeSnapshot', name='bar')
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class VolumeTypeConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(VolumeTypeConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_volume_type = mock.Mock()
        self.ctx.clients.client_plugin(
            'cinder').get_volume_type = self.mock_get_volume_type
        self.constraint = cinder.VolumeTypeConstraint()

    def test_validation(self):
        self.mock_get_volume_type.return_value = 'volume_type'
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_volume_type.side_effect = exception.EntityNotFound(
            entity='VolumeType', name='bar')
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class VolumeBackupConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(VolumeBackupConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_volume_backup = mock.Mock()
        self.ctx.clients.client_plugin(
            'cinder').get_volume_backup = self.mock_get_volume_backup
        self.constraint = cinder.VolumeBackupConstraint()

    def test_validation(self):
        self.mock_get_volume_backup.return_value = 'volume_backup'
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        ex = exception.EntityNotFound(entity='Volume backup', name='bar')
        self.mock_get_volume_backup.side_effect = ex
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class QoSSpecsConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(QoSSpecsConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_qos_specs = mock.Mock()
        self.ctx.clients.client_plugin(
            'cinder').get_qos_specs = self.mock_get_qos_specs
        self.constraint = cinder.QoSSpecsConstraint()

    def test_validation(self):
        self.mock_get_qos_specs.return_value = None
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_qos_specs.side_effect = cinder_exc.NotFound(404)
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class CinderClientAPIVersionTest(common.HeatTestCase):

    def test_cinder_api_v3(self):
        ctx = utils.dummy_context()
        self.patchobject(ctx.keystone_session, 'get_endpoint')
        ctx.clients.client_plugin(
            'cinder').max_microversion = cinder_api_versions.MAX_VERSION
        client = ctx.clients.client('cinder')
        self.assertEqual('3.0', client.version)

    def test_cinder_api_not_supported(self):
        ctx = utils.dummy_context()
        self.patchobject(ctx.keystone_session, 'get_endpoint',
                         side_effect=[ks_exceptions.EndpointNotFound,
                                      ks_exceptions.EndpointNotFound])
        self.assertRaises(exception.Error, ctx.clients.client, 'cinder')


class CinderClientPluginExtensionsTest(CinderClientPluginTest):
    """Tests for extensions in cinderclient."""

    def test_has_no_extensions(self):
        self.cinder_client.list_extensions.show_all.return_value = []
        self.assertFalse(self.cinder_plugin.has_extension(
            "encryption"))

    def test_has_no_interface_extensions(self):
        mock_extension = mock.Mock()
        p = mock.PropertyMock(return_value='os-xxxx')
        type(mock_extension).alias = p
        self.cinder_client.list_extensions.show_all.return_value = [
            mock_extension]
        self.assertFalse(self.cinder_plugin.has_extension(
            "encryption"))

    def test_has_os_interface_extension(self):
        mock_extension = mock.Mock()
        p = mock.PropertyMock(return_value='encryption')
        type(mock_extension).alias = p
        self.cinder_client.list_extensions.show_all.return_value = [
            mock_extension]
        self.assertTrue(self.cinder_plugin.has_extension(
            "encryption"))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from keystoneauth1 import loading as ks_loading
from keystoneauth1 import session

from heat.common import auth_plugin
from heat.common import config
from heat.common import exception
from heat.common import policy
from heat.tests import common


class TestAuthPlugin(common.HeatTestCase):

    def setUp(self):
        self.credential = (
            '{"auth_type": "v3applicationcredential", '
            '"auth": {"auth_url": "http://192.168.1.101/identity/v3", '
            '"application_credential_id": '
            '"9dfa187e5a354484bf9c49a2b674333a", '
            '"application_credential_secret": "sec"} }')
        self.m_plugin = mock.Mock()
        self.m_loader = self.patchobject(
            ks_loading, 'get_plugin_loader', return_value=self.m_plugin)
        self.patchobject(policy.Enforcer, 'check_is_admin')
        self.secret_id = '0eca0615-c330-41aa-b0cb-a2493a770409'
        self.session = session.Session(
            **config.get_ssl_options('keystone'))
        super(TestAuthPlugin, self).setUp()

    def _get_keystone_plugin_loader(self):
        auth_plugin.get_keystone_plugin_loader(self.credential, self.session)

        self.m_plugin.load_from_options.assert_called_once_with(
            application_credential_id='9dfa187e5a354484bf9c49a2b674333a',
            application_credential_secret='sec',
            auth_url='http://192.168.1.101/identity/v3')
        self.m_loader.assert_called_once_with('v3applicationcredential')

    def test_get_keystone_plugin_loader(self):
        self._get_keystone_plugin_loader()
        # called in validate_auth_plugin
        self.assertEqual(
            1, self.m_plugin.load_from_options().get_token.call_count)

    def test_get_keystone_plugin_loader_with_no_AuthZ(self):
        self.m_plugin.load_from_options().get_token.side_effect = Exception
        self.assertRaises(
            exception.AuthorizationFailure, self._get_keystone_plugin_loader)
        self.assertEqual(
            1, self.m_plugin.load_from_options().get_token.call_count)

    def test_parse_auth_credential_to_dict(self):
        cred_dict = json.loads(self.credential)
        self.assertEqual(
            cred_dict, auth_plugin.parse_auth_credential_to_dict(
                self.credential))

    def test_parse_auth_credential_to_dict_with_value_error(self):
        credential = (
            '{"auth": {"auth_url": "http://192.168.1.101/identity/v3", '
            '"application_credential_id": '
            '"9dfa187e5a354484bf9c49a2b674333a", '
            '"application_credential_secret": "sec"} }')
        error = self.assertRaises(
            ValueError, auth_plugin.parse_auth_credential_to_dict, credential)
        self.assertEqual("Missing key in auth information, the correct "
                         "format contains [\'auth_type\', \'auth\'].",
                         str(error))

    def test_parse_auth_credential_to_dict_with_json_error(self):
        credential = (
            "{'auth_type': v3applicationcredential, "
            "'auth': {'auth_url': 'http://192.168.1.101/identity/v3', "
            "'application_credential_id': "
            "'9dfa187e5a354484bf9c49a2b674333a', "
            "'application_credential_secret': 'sec'} }")
        error = self.assertRaises(
            ValueError, auth_plugin.parse_auth_credential_to_dict, credential)
        error_msg = ('Failed to parse credential, please check your Stack '
                     'Credential format.')
        self.assertEqual(error_msg, str(error))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from urllib import parse

from heat.engine.clients.os.keystone import fake_keystoneclient as fake_ks
from heat.engine import resource
from heat.engine.resources.openstack.keystone import region
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

KEYSTONE_REGION_TEMPLATE = {
    'heat_template_version': '2015-10-15',
    'resources': {
        'test_region': {
            'type': 'OS::Keystone::Region',
            'properties': {
                'id': 'test_region_1',
                'description': 'Test region',
                'parent_region': 'default_region',
                'enabled': 'True'
            }
        }
    }
}


class KeystoneRegionTest(common.HeatTestCase):
    def setUp(self):
        super(KeystoneRegionTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack_keystone',
            template.Template(KEYSTONE_REGION_TEMPLATE)
        )

        self.test_region = self.stack['test_region']

        # Mock client
        self.keystoneclient = mock.Mock()
        self.patchobject(resource.Resource, 'client',
                         return_value=fake_ks.FakeKeystoneClient(
                             client=self.keystoneclient))
        self.regions = self.keystoneclient.regions

        keystone_client_plugin = mock.MagicMock()
        self.test_region.client_plugin = mock.MagicMock()
        self.test_region.client_plugin.return_value = keystone_client_plugin

    def _get_mock_region(self):
        value = mock.MagicMock()
        region_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        value.id = region_id

        return value

    def test_region_handle_create(self):
        mock_region = self._get_mock_region()
        self.regions.create.return_value = mock_region

        # validate the properties
        self.assertEqual(
            'test_region_1',
            self.test_region.properties.get(region.KeystoneRegion.ID))
        self.assertEqual(
            'Test region',
            self.test_region.properties.get(
                region.KeystoneRegion.DESCRIPTION))
        self.assertEqual(
            'default_region',
            self.test_region.properties.get(
                region.KeystoneRegion.PARENT_REGION))
        self.assertEqual(
            True,
            self.test_region.properties.get(region.KeystoneRegion.ENABLED))

        self.test_region.handle_create()

        # validate region creation
        self.regions.create.assert_called_once_with(
            id=parse.quote('test_region_1'),
            description='Test region',
            parent_region='default_region',
            enabled=True)

        # validate physical resource id
        self.assertEqual(mock_region.id, self.test_region.resource_id)

    def test_region_handle_create_minimal(self):
        values = {
            'description': 'sample region',
            'enabled': True,
            'parent_region': None,
            'id': None
        }

        def _side_effect(key):
            return values[key]

        mock_region = self._get_mock_region()
        self.regions.create.return_value = mock_region
        self.test_region.properties = mock.MagicMock()
        self.test_region.properties.__getitem__.side_effect = _side_effect

        self.test_region.handle_create()

        self.regions.create.assert_called_once_with(
            id=None,
            description='sample region',
            parent_region=None,
            enabled=True)

    def test_region_handle_update(self):
        self.test_region.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {region.KeystoneRegion.DESCRIPTION:
                     'Test Region updated',
                     region.KeystoneRegion.ENABLED: False,
                     region.KeystoneRegion.PARENT_REGION: 'test_parent_region'}

        self.test_region.handle_update(json_snippet=None,
                                       tmpl_diff=None,
                                       prop_diff=prop_diff)

        self.regions.update.assert_called_once_with(
            region=self.test_region.resource_id,
            description=prop_diff[region.KeystoneRegion.DESCRIPTION],
            enabled=prop_diff[region.KeystoneRegion.ENABLED],
            parent_region='test_parent_region'
        )

    def test_region_get_live_state(self):
        self.test_region.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_dict = mock.MagicMock()
        mock_dict.to_dict.return_value = {
            "parent_region_id": None,
            "enabled": True,
            "id": "79e4d02f8b454a7885c413d5d4297813",
            "links": {"self": "link"},
            "description": ""
        }
        self.regions.get.return_value = mock_dict

        reality = self.test_region.get_live_state(self.test_region.properties)
        expected = {
            "parent_region": None,
            "enabled": True,
            "description": ""
        }
        self.assertEqual(expected, reality)



# Copyright 2017 Ericsson
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation

LOG = logging.getLogger(__name__)


class Trunk(neutron.NeutronResource):
    """A resource for managing Neutron trunks.

    Requires Neutron Trunk Extension to be enabled::

      $ openstack extension show trunk

    The network trunk service allows multiple networks to be connected to
    an instance using a single virtual NIC (vNIC). Multiple networks can
    be presented to an instance by connecting the instance to a single port.

    Users can create a port, associate it with a trunk (as the trunk's parent)
    and launch an instance on that port. Users can dynamically attach and
    detach additional networks without disrupting operation of the instance.

    Every trunk has a parent port and can have any number (0, 1, ...) of
    subports. The parent port is the port that the instance is directly
    associated with and its traffic is always untagged inside the instance.
    Users must specify the parent port of the trunk when launching an
    instance attached to a trunk.

    A network presented by a subport is the network of the associated port.
    When creating a subport, a ``segmentation_type`` and ``segmentation_id``
    may be required by the driver so the user can distinguish the networks
    inside the instance. As of release Pike only ``segmentation_type``
    ``vlan`` is supported. ``segmentation_id`` defines the segmentation ID
    on which the subport network is presented to the instance.

    Note that some Neutron backends (primarily Open vSwitch) only allow
    trunk creation before an instance is booted on the parent port. To avoid
    a possible race condition when booting an instance with a trunk it is
    strongly recommended to refer to the trunk's parent port indirectly in
    the template via ``get_attr``. For example::

      trunk:
        type: OS::Neutron::Trunk
        properties:
          port: ...
      instance:
        type: OS::Nova::Server
        properties:
          networks:
            - { port: { get_attr: [trunk, port_id] } }

    Though other Neutron backends may tolerate the direct port reference
    (and the possible reverse ordering of API requests implied) it's a good
    idea to avoid writing Neutron backend specific templates.
    """

    entity = 'trunk'

    required_service_extension = 'trunk'

    support_status = support.SupportStatus(
        status=support.SUPPORTED,
        version='9.0.0',
    )

    PROPERTIES = (
        NAME, PARENT_PORT, SUB_PORTS, DESCRIPTION, ADMIN_STATE_UP,
    ) = (
        'name', 'port', 'sub_ports', 'description', 'admin_state_up',
    )

    _SUBPORT_KEYS = (
        PORT, SEGMENTATION_TYPE, SEGMENTATION_ID,
    ) = (
        'port', 'segmentation_type', 'segmentation_id',
    )

    _subport_schema = {
        PORT: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of a port to be used as a subport.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.port'),
            ],
        ),
        SEGMENTATION_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Segmentation type to be used on the subport.'),
            required=True,
            # TODO(nilles): custom constraint 'neutron.trunk_segmentation_type'
            constraints=[
                constraints.AllowedValues(['vlan']),
            ],
        ),
        SEGMENTATION_ID: properties.Schema(
            properties.Schema.INTEGER,
            _('The segmentation ID on which the subport network is presented '
              'to the instance.'),
            required=True,
            # TODO(nilles): custom constraint 'neutron.trunk_segmentation_id'
            constraints=[
                constraints.Range(1, 4094),
            ],
        ),
    }

    ATTRIBUTES = (
        PORT_ATTR,
    ) = (
        'port_id',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('A string specifying a symbolic name for the trunk, which is '
              'not required to be uniqe.'),
            update_allowed=True,
        ),
        PARENT_PORT: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of a port to be used as a parent port.'),
            required=True,
            immutable=True,
            constraints=[
                constraints.CustomConstraint('neutron.port'),
            ],
        ),
        SUB_PORTS: properties.Schema(
            properties.Schema.LIST,
            _('List with 0 or more map elements containing subport details.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema=_subport_schema,
            ),
            update_allowed=True,
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description for the trunk.'),
            update_allowed=True,
        ),
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Enable/disable subport addition, removal and trunk delete.'),
            update_allowed=True,
        ),
    }

    attributes_schema = {
        PORT_ATTR: attributes.Schema(
            _('ID or name of a port used as a parent port.'),
            type=attributes.Schema.STRING,
        ),
    }

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.PARENT_PORT],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_PORT
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.SUB_PORTS, self.PORT],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_PORT
            ),
        ]

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        props['port_id'] = props.pop(self.PARENT_PORT)

        if self.SUB_PORTS in props and props[self.SUB_PORTS]:
            for sub_port in props[self.SUB_PORTS]:
                sub_port['port_id'] = sub_port.pop(self.PORT)

        LOG.debug('attempt to create trunk: %s', props)
        trunk = self.client().create_trunk({'trunk': props})['trunk']
        self.resource_id_set(trunk['id'])

    def check_create_complete(self, *args):
        attributes = self._show_resource()
        return self.is_built(attributes)

    def handle_delete(self):
        if self.resource_id is not None:
            with self.client_plugin().ignore_not_found:
                LOG.debug('attempt to delete trunk: %s', self.resource_id)
                self.client().delete_trunk(self.resource_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        """Handle update to a trunk in (at most) three neutron calls.

        Call #1) Update all changed properties but 'sub_ports'.
            PUT /v2.0/trunks/TRUNK_ID
            openstack network trunk set

        Call #2) Delete subports not needed anymore.
            PUT /v2.0/trunks/TRUNK_ID/remove_subports
            openstack network trunk unset --subport

        Call #3) Create new subports.
            PUT /v2.0/trunks/TRUNK_ID/add_subports
            openstack network trunk set --subport

        A single neutron port cannot be two subports at the same time (ie.
        have two segmentation (type, ID)s on the same trunk or to belong to
        two trunks). Therefore we have to delete old subports before creating
        new ones to avoid conflicts.
        """

        LOG.debug('attempt to update trunk %s', self.resource_id)

        # NOTE(bence romsics): We want to do set operations on the subports,
        # however we receive subports represented as dicts. In Python
        # mutable objects like dicts are not hashable so they cannot be
        # inserted into sets. So we convert subport dicts to (immutable)
        # frozensets in order to do the set operations.
        def dict2frozenset(d):
            """Convert a dict to a frozenset.

            Create an immutable equivalent of a dict, so it's hashable
            therefore can be used as an element of a set or a key of another
            dictionary.
            """
            return frozenset(d.items())

        # NOTE(bence romsics): prop_diff contains a shallow diff of the
        # properties, so if we had used that to update subports we would
        # re-create all subports even if just a single subport changed. So we
        # need to forget about prop_diff['sub_ports'] and diff out the real
        # subport changes from self.properties and json_snippet.
        if 'sub_ports' in prop_diff:
            del prop_diff['sub_ports']

        sub_ports_prop_old = self.properties[self.SUB_PORTS] or []
        sub_ports_prop_new = json_snippet.properties(
            self.properties_schema)[self.SUB_PORTS] or []

        subports_old = {dict2frozenset(d): d for d in sub_ports_prop_old}
        subports_new = {dict2frozenset(d): d for d in sub_ports_prop_new}

        old_set = set(subports_old.keys())
        new_set = set(subports_new.keys())

        delete = old_set - new_set
        create = new_set - old_set

        dicts_delete = [subports_old[fs] for fs in delete]
        dicts_create = [subports_new[fs] for fs in create]

        LOG.debug('attempt to delete subports of trunk %s: %s',
                  self.resource_id, dicts_delete)
        LOG.debug('attempt to create subports of trunk %s: %s',
                  self.resource_id, dicts_create)

        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client().update_trunk(self.resource_id, {'trunk': prop_diff})

        if dicts_delete:
            delete_body = self.prepare_trunk_remove_subports_body(dicts_delete)
            self.client().trunk_remove_subports(self.resource_id, delete_body)

        if dicts_create:
            create_body = self.prepare_trunk_add_subports_body(dicts_create)
            self.client().trunk_add_subports(self.resource_id, create_body)

    def check_update_complete(self, *args):
        attributes = self._show_resource()
        return self.is_built(attributes)

    @staticmethod
    def prepare_trunk_remove_subports_body(subports):
        """Prepares body for PUT /v2.0/trunks/TRUNK_ID/remove_subports."""

        return {
            'sub_ports': [
                {'port_id': sp['port']} for sp in subports
            ]
        }

    @staticmethod
    def prepare_trunk_add_subports_body(subports):
        """Prepares body for PUT /v2.0/trunks/TRUNK_ID/add_subports."""

        return {
            'sub_ports': [
                {'port_id': sp['port'],
                 'segmentation_type': sp['segmentation_type'],
                 'segmentation_id': sp['segmentation_id']}
                for sp in subports
            ]
        }


def resource_mapping():
    return {
        'OS::Neutron::Trunk': Trunk,
    }
